/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 1999-2001 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */

#define  __NO_VERSION__

#include "nv-misc.h"
#include "os-interface.h"
#include "nv-linux.h"

#if defined(NV_LINUX_ACPI_EVENTS_SUPPORTED)

/*
 * This code allows RM to subscribe to events surrounding the power adapter.
 *
 */

/*
 * c file definitions of function that do not need external visibility
 *
 */
static int         nv_acpi_add         (struct acpi_device *device);
static int         nv_acpi_remove      (struct acpi_device *device, int type);
static void        nv_acpi_event       (acpi_handle, u32, void *);
static int         nv_acpi_match       (struct acpi_device *, struct acpi_driver *);

#ifdef DEBUG
/*
 * internal debug routines, very useful for viewing the namespace of the VGA device
 */
static void        nv_acpi_dump_device (acpi_handle handle);
static acpi_status nv_acpi_walk_callback (acpi_handle, u32, void *, void **);
#endif //DEBUG

static struct acpi_driver nv_acpi_driver = {
    .name = "NVIDIA ACPI Video Driver",
    .class = "video",
    .ops = {
        .add = nv_acpi_add,
        .remove = nv_acpi_remove,
        .match = nv_acpi_match,
    },
};

int nv_acpi_init(nv_state_t *nv)
{
    /*
     * this function will register us with the Linux
     * ACPI subsystem.
     */

    acpi_status status;

    // return silent if we are already in place
    if (nv->acpi_initialized)
        return 0;

    // register with the OS for VGA devices
    status = acpi_bus_register_driver(&nv_acpi_driver);

    if (ACPI_FAILURE(status) < 0)
    {
        nv_printf(NV_DBG_INFO,
            "NVRM: nv_acpi_init: acpi_bus_register_driver() failed (%d)!\n", status);
    }
    else
    {
        // mark ourselves as initialized
        nv->acpi_initialized = 1;
    }

    return status;
}

int nv_acpi_uninit(nv_state_t *nv)
{
    // return clean if we are not initialized
    if (!nv->acpi_initialized)
        return 0;

    acpi_bus_unregister_driver(&nv_acpi_driver);

    // clear initialization flag
    nv->acpi_initialized = 0;

    return 0;
}

static int nv_acpi_add(struct acpi_device *device)
{
    /*
     * This function will cause RM to initialize the things it needs for acpi interaction
     * on the display device.
     */
    int status = -1;
    RM_STATUS rmStatus = RM_ERROR;
    nv_acpi_t *pNvAcpiObject = NULL;
    union acpi_object control_argument_0 = { ACPI_TYPE_INTEGER };    
    struct acpi_object_list control_argument_list = { 0, NULL };
    nv_stack_t *sp = NULL;

    NV_KMEM_CACHE_ALLOC_STACK(sp);
    if (sp == NULL)
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: failed to allocate stack!\n");
        return -ENOMEM;
    }

    // allocate data structure we need
    rmStatus = os_alloc_mem((void **) &pNvAcpiObject, sizeof(nv_acpi_t));
    if (rmStatus != RM_OK)
    {
        NV_KMEM_CACHE_FREE_STACK(sp);
        nv_printf(NV_DBG_ERRORS,
            "NVRM: nv_acpi_add: failed to allocate ACPI device management data!\n");
        return -ENOMEM;
    }

    os_mem_set((void *)pNvAcpiObject, 0, sizeof(nv_acpi_t));

    // assign driver data structure ptr to this device
    acpi_driver_data(device) = pNvAcpiObject;

    // store a device reference in our object
    pNvAcpiObject->device = device;
    pNvAcpiObject->sp = sp;

    // grab handles to all the important nodes representing devices
    status = acpi_get_handle(device->handle, "CRT", &pNvAcpiObject->crt_handle);

    if (ACPI_FAILURE(status))
    {
        nv_printf(NV_DBG_INFO,
            "NVRM: nv_acpi_add: failed to retrieve ACPI \"CRT\" node handle.\n");
    }

    status = acpi_get_handle(device->handle, "LCD", &pNvAcpiObject->lcd_handle);

    if (ACPI_FAILURE(status))
    {
        nv_printf(NV_DBG_INFO,
            "NVRM: nv_acpi_add: failed to retrieve ACPI \"LCD\" node handle.\n");
    }

    status = acpi_get_handle(device->handle, "TV",  &pNvAcpiObject->tv_handle );

    if (ACPI_FAILURE(status))
    {
        nv_printf(NV_DBG_INFO,
            "NVRM: nv_acpi_add: failed to retrieve ACPI \"TV\" node handle.\n");
    }

    status = acpi_get_handle(device->handle, "DVI", &pNvAcpiObject->dvi_handle);

    if (ACPI_FAILURE(status))
    {
        nv_printf(NV_DBG_INFO,
            "NVRM: nv_acpi_add: failed to retrieve ACPI \"DVI\" node handle.\n");
    }

    // arg 0, bits 1:0, 0 = enable events
    control_argument_0.integer.type = ACPI_TYPE_INTEGER;
    control_argument_0.integer.value = 0x0;

    // listify it
    control_argument_list.count = 1;
    control_argument_list.pointer = &control_argument_0;

    // _DOS method takes 1 argument and returns nothing
    status = acpi_evaluate_object(device->handle, "_DOS", &control_argument_list, NULL);

    if (ACPI_FAILURE(status))
    {
        nv_printf(NV_DBG_INFO,
            "NVRM: nv_acpi_add: failed to enable display switch events (%d)!\n", status);
    }

    status = acpi_install_notify_handler(device->handle, ACPI_DEVICE_NOTIFY,
                    nv_acpi_event, pNvAcpiObject);

    if (ACPI_FAILURE(status))
    {
        nv_printf(NV_DBG_INFO,
            "NVRM: nv_acpi_add: failed to install event notification handler (%d)!\n", status);
    }
    else
    {
        try_module_get(THIS_MODULE);
        pNvAcpiObject->notify_handler_installed = 1;
    }

    return 0;
}

static int nv_acpi_remove(struct acpi_device *device, int type)
{
    /*
     * This function will cause RM to relinquish control of the VGA ACPI device.
     */
    acpi_status status;
    union acpi_object control_argument_0 = { ACPI_TYPE_INTEGER };    
    struct acpi_object_list control_argument_list = { 0, NULL };
    nv_acpi_t *pNvAcpiObject = acpi_driver_data(device);

    // arg 0, bits 1:0, 1 = disable events
    control_argument_0.integer.type = ACPI_TYPE_INTEGER;
    control_argument_0.integer.value = 0x1;

    // listify it
    control_argument_list.count = 1;
    control_argument_list.pointer = &control_argument_0;

    // _DOS method takes 1 argument and returns nothing
    status = acpi_evaluate_object(device->handle, "_DOS", &control_argument_list, NULL);

    if (ACPI_FAILURE(status))
    {
        nv_printf(NV_DBG_INFO,
            "NVRM: nv_acpi_remove: failed to disable display switch events (%d)!\n", status);
    }

    if (pNvAcpiObject->notify_handler_installed)
    {
        // no status returned for this function
        acpi_os_wait_events_complete(NULL);

        // remove event notifier
        status = acpi_remove_notify_handler(device->handle, ACPI_DEVICE_NOTIFY, nv_acpi_event);
    }

    if (pNvAcpiObject->notify_handler_installed &&
        ACPI_FAILURE(status))
    {
        nv_printf(NV_DBG_INFO,
            "NVRM: nv_acpi_remove: failed to remove event notification handler (%d)!\n", status);
    }
    else
    {
        NV_KMEM_CACHE_FREE_STACK(pNvAcpiObject->sp);
        os_free_mem((void *)pNvAcpiObject);
        module_put(THIS_MODULE);
    }

    return status;
}

static void nv_acpi_event(acpi_handle handle, u32 event, void *data)
{
    /*
     * This function will handle acpi events from the linux kernel, used
     * to detect notifications from the VGA device.
     */
    nv_acpi_t *pNvAcpiObject = data;

    nv_printf(NV_DBG_INFO, "NVRM: nv_acpi_event: Event 0x%x\n", event);

    rm_system_event(pNvAcpiObject->sp, event);

    // no unsubscription or re-enable necessary. Once DOD has been set, we are go.
    // once we are subscribed to ACPI events, we don't have to re-subscribe unless
    // unsubscribe.
}

static int nv_acpi_match(struct acpi_device *device, struct acpi_driver *driver)
{
    acpi_handle DOS_method_handler, DOD_method_handler;

    /*
     *
     * attempt to get handles to all of these objects
     * _DOS = enable/disable output switching method
     * _DOD = enumerate all devices attached to the display driver
     * 
     * If handles are available, this is the VGA object.
     *
     */

    if ((!acpi_get_handle(device->handle, "_DOS", &DOS_method_handler)) &&
        (!acpi_get_handle(device->handle, "_DOD", &DOD_method_handler)))
    {
        nv_printf(NV_DBG_INFO, "NVRM: nv_acpi_match: VGA ACPI device detected.\n");

        return 0;
    }
    else
        return (-ENODEV);
}

#ifdef DEBUG
#define NV_ACPI_NAME_BUFFER_LENGTH 1024

/*
 * utility function, not normally used except for debugging.
 */
acpi_status nv_acpi_walk_callback(
    acpi_handle handle,
    u32 nest_level,
    void *acpi_name,
    void **dummy2
)
{
    acpi_status status;
    struct acpi_buffer result_buffer;
    char *name = acpi_name;

    result_buffer.length = NV_ACPI_NAME_BUFFER_LENGTH;
    result_buffer.pointer = acpi_name;
    
    status = acpi_get_name(handle, ACPI_SINGLE_NAME, &result_buffer);

    if (ACPI_SUCCESS(status))
    {
        nv_printf(NV_DBG_ERRORS,
            "NVRM: nv_acpi_walk_callback: level: %d, name: %s\n", nest_level, name);
    }

    return 0;
}

/*
 * utility function, not normally used except for debugging.
 */
static void nv_acpi_dump_device(acpi_handle handle)
{
    acpi_status status;
    struct acpi_buffer result_buffer;
    void *acpi_name;

    if (os_alloc_mem(&acpi_name, NV_ACPI_NAME_BUFFER_LENGTH) != RM_OK)
    {
        nv_printf(NV_DBG_ERRORS,
            "NVRM: nv_acpi_dump_device: failed to allocate ACPI name buffer!\n");
        return;
    }

    result_buffer.length = NV_ACPI_NAME_BUFFER_LENGTH;
    result_buffer.pointer = acpi_name;

    status = acpi_get_name(handle, ACPI_SINGLE_NAME, &result_buffer);

    if (ACPI_SUCCESS(status))
    {
        nv_printf(NV_DBG_ERRORS,
            "NVRM: nv_acpi_dump_device: Preparing to print namespace for %s\n",
            (char *)acpi_name);
    }
    else
    {
        nv_printf(NV_DBG_ERRORS,
            "NVRM: nv_acpi_dump_device: Preparing to print namespace for "
            "NAMESPACE INVALID\n");
    }

    acpi_walk_namespace(ACPI_TYPE_ANY, handle, 1, nv_acpi_walk_callback, acpi_name, NULL);

    os_free_mem(acpi_name);
}
#endif // DEBUG

#else // NV_LINUX_ACPI_EVENTS_SUPPORTED

int nv_acpi_init(nv_state_t *nv)
{
    return 0;
}

int nv_acpi_uninit(nv_state_t *nv)
{
    return 0;
}

#endif
